DeclareModule dPort

Enumeration
  #PBIO_OK
  #PBIO_ErrCopyDriver
  #PBIO_ErrInstallDriver
  #PBIO_ErrAccess
EndEnumeration

Declare dpInit()
Declare dpClose()
Declare dpOpenDriver()
Declare dpInp(PortAddress.u)
Declare dpOut(PortAddress.u, Byte.a)
Declare dpStatus()
Declare dpID()
Declare dpInstall()
Declare dpUnInstall()
Declare OSbits()
EndDeclareModule

Module dPort
#IOCTL_READ_PORT_UCHAR   = $200
#IOCTL_WRITE_PORT_UCHAR  = $400

Structure Param_Info
  hDriverID.i
  ErrStatus.i
  DrFileName.s
  DriverName.s
  *Point
  Len.i
  IsWin9x.a
EndStructure

Prototype pIsWow64Process(hProcess.i, *Wow64Process)

EnableExplicit
  
Global ParamInfo.Param_Info

;    .
Procedure.a InPort(Port.u)
  Protected Result.a=0
  EnableASM
  MOV DX, Port
  IN  al,DX
  MOV Result, al
  DisableASM
  ProcedureReturn Result
EndProcedure

;    .
Procedure OutPort(Port.u, Byte.a)
  EnableASM
  MOV al, Byte
  MOV DX, Port
  OUT DX, al
  DisableASM
EndProcedure

Procedure OSbits()
  Protected Is64BitOS = 0 ; Returns 1 for 64bit, else 0.
  Protected hDLL, IsWow64Process
  
  If ParamInfo\IsWin9x=#True ; Win9x   64 .
    Is64BitOS = 0
  Else
    
    CompilerIf #PB_Compiler_Processor = #PB_Processor_x64
      Is64BitOS=1
    CompilerElse
      
      Is64BitOS=0
      hDll = OpenLibrary(#PB_Any, "kernel32.dll")
      If hDll
        
        IsWow64Process.pIsWow64Process = GetFunction(hDll,"IsWow64Process")
        If IsWow64Process
          IsWow64Process(GetCurrentProcess_(), @Is64BitOS)
        EndIf
        
        CloseLibrary(hDll)
      EndIf
      
    CompilerEndIf
    
  EndIf
  
  ProcedureReturn Is64BitOS
EndProcedure

Procedure Driver_Start()
  Protected hSCManager.i, hServ.i
  Protected ErrorCode, Result, Drivers_Name.s
  
  Drivers_Name = GetFilePart(ParamInfo\DriverName, #PB_FileSystem_NoExtension)
  
  Result=#False
  hSCManager  = OpenSCManager_(#Null, #Null, #SC_MANAGER_ALL_ACCESS)
  If hSCManager = 0
    If GetLastError_() = #ERROR_ACCESS_DENIED
      hSCManager  = OpenSCManager_(#Null, #Null, #GENERIC_READ)
      hServ       = OpenService_(hSCManager, @Drivers_Name, #GENERIC_EXECUTE)
      If StartService_(hServ, 0, #Null)
        Result=#True
      EndIf
    EndIf
    
  Else
    hServ       = OpenService_(hSCManager, @Drivers_Name, #GENERIC_EXECUTE)
    If StartService_(hServ, 0, #Null)
      Result=#True  
    Else
      ErrorCode=GetLastError_()
      If ErrorCode=1056
        Result=#True
      EndIf
    EndIf
  EndIf
  
  If hServ : CloseServiceHandle_(hServ) : EndIf
  If hSCManager : CloseServiceHandle_(hSCManager) : EndIf
  
  ProcedureReturn Result
EndProcedure

Procedure dpOpenDriver()
  Protected s.s
  
  ParamInfo\ErrStatus = #PBIO_OK
  
  If ParamInfo\IsWin9x=#True
    ParamInfo\hDriverID=2
  Else
    
    If ParamInfo\hDriverID
      CloseHandle_(ParamInfo\hDriverID)
      ParamInfo\hDriverID=0
    EndIf
    
    s="\\.\"+ParamInfo\DriverName
    
    ParamInfo\hDriverID = CreateFile_(s, #GENERIC_READ | #GENERIC_WRITE, 0, 0, #OPEN_EXISTING, #FILE_ATTRIBUTE_NORMAL, 0)
    If ParamInfo\hDriverID=#INVALID_HANDLE_VALUE
      If Driver_Start()
        ParamInfo\hDriverID = CreateFile_(s, #GENERIC_READ | #GENERIC_WRITE, 0, 0, #OPEN_EXISTING, #FILE_ATTRIBUTE_NORMAL, 0)
      EndIf
    EndIf
    
    If ParamInfo\hDriverID = #INVALID_HANDLE_VALUE
      ParamInfo\hDriverID = 0
    EndIf
    
    If ParamInfo\hDriverID=0 And ParamInfo\ErrStatus = #PBIO_OK
      ParamInfo\ErrStatus = #PBIO_ErrAccess
    EndIf
    
  EndIf
  
  ProcedureReturn ParamInfo\hDriverID
EndProcedure

Procedure dpInit()
  Protected OS_Version
  
  OS_Version = OSVersion()
    
  If OS_Version=#PB_OS_Windows_95 Or OS_Version=#PB_OS_Windows_98 Or OS_Version=#PB_OS_Windows_ME
    ParamInfo\IsWin9x=#True ;   Windows9x.
    ParamInfo\hDriverID=2
  Else
    
    ParamInfo\IsWin9x=#False
        
    ParamInfo\ErrStatus=#PBIO_OK
    ParamInfo\DrFileName = GetPathPart(ProgramFilename())+"DriverPortIO"
    ParamInfo\DriverName = "DrPortIO_28"
    
    If OSbits() ; x64.
      ParamInfo\DrFileName+"_x64"
      ParamInfo\DriverName+"_x64"
      ParamInfo\Point = ?Driver_x64
      ParamInfo\Len   = ?DriverEnd_x64-?Driver_x64
    Else        ; x86.
      ParamInfo\Point = ?Driver
      ParamInfo\Len   = ?DriverEnd-?Driver
    EndIf
    ParamInfo\DrFileName+".sys"
    
    ParamInfo\hDriverID = dpOpenDriver()
    
  EndIf
  
  ProcedureReturn ParamInfo\hDriverID
EndProcedure

Procedure dpClose()
  If ParamInfo\hDriverID And ParamInfo\IsWin9x<>#True
    CloseHandle_(ParamInfo\hDriverID)
  EndIf
  ParamInfo\hDriverID=0
EndProcedure


Procedure dpInp(PortAddress.u) ;    ,     'PortAddress'.
  Protected Result, Buffer.i, BytesReturned.i, OutBuff.i
  Result=0 : Buffer=0 : OutBuff=0
  
  If ParamInfo\IsWin9x=#True
    Result=InPort(PortAddress)
  Else
    
    If ParamInfo\hDriverID
      Buffer=PortAddress
      DeviceIoControl_(ParamInfo\hDriverID, #IOCTL_READ_PORT_UCHAR, @Buffer, 
                       SizeOf(Buffer), @OutBuff, SizeOf(OutBuff), @BytesReturned, 0)
      Result=OutBuff
    EndIf
    
  EndIf
  
  ProcedureReturn Result
EndProcedure

Procedure dpOut(PortAddress.u, Byte.a) ;    ,     'PortAddress'.
  Protected Result, Buffer.i, BytesReturned.i, OutBuff.i
  Result=#False
  
  If ParamInfo\IsWin9x=#True
    OutPort(PortAddress, Byte)
  Else
    
    Result=#True
    If ParamInfo\hDriverID
      Buffer=PortAddress
      PokeA(@Buffer+2, Byte)
      If DeviceIoControl_(ParamInfo\hDriverID, #IOCTL_WRITE_PORT_UCHAR, @Buffer, 
                          SizeOf(Buffer), @OutBuff, SizeOf(OutBuff), @BytesReturned, 0)
        Result = #True
      EndIf
    EndIf
    
  EndIf
  
  ProcedureReturn Result
EndProcedure


Procedure dpStatus() ;  .   ,  0.
  ProcedureReturn ParamInfo\ErrStatus
EndProcedure

Procedure dpID()
  ProcedureReturn ParamInfo\hDriverID
EndProcedure

Procedure dpUnInstall()
  Protected hSCManager.i, hServ.i, Drivers_Name.s
  Protected ServiceStatus.SERVICE_STATUS, Result=#False
  
  If ParamInfo\IsWin9x=#True
    ProcedureReturn #True
  EndIf
  
  Drivers_Name=GetFilePart(ParamInfo\DriverName, #PB_FileSystem_NoExtension)
  hSCManager  = OpenSCManager_(#Null, #Null, #SC_MANAGER_ALL_ACCESS)
  hServ       = OpenService_(hSCManager, @Drivers_Name, #SERVICE_ALL_ACCESS)
  
  If hServ<>#Null
     ControlService_(hServ, #SERVICE_CONTROL_STOP, @ServiceStatus)
     If DeleteService_(hServ)
       Result=#True
     EndIf
  EndIf
  
  If hServ : CloseServiceHandle_(hServ) : EndIf
  If hSCManager : CloseServiceHandle_(hSCManager) : EndIf
  
  
  ProcedureReturn Result
EndProcedure

Procedure dpInstall()
  Protected Result, FileID, File.s, Mgr
  Protected ErrorCode, Ser, Count, Drivers_Name.s
  Result=#False
  Count=0 : FileID=0 : Mgr=0 : Ser=0
  
  If ParamInfo\IsWin9x=#True
    ProcedureReturn #True
  EndIf
  
  ParamInfo\ErrStatus = #PBIO_OK
  
  File=ParamInfo\DrFileName
  Drivers_Name=GetFilePart(ParamInfo\DriverName, #PB_FileSystem_NoExtension)
  
  If FileSize(File)<=0
    
    FileID=CreateFile(#PB_Any, File)
    If FileID
      WriteData(FileID, ParamInfo\Point, ParamInfo\Len)
      CloseFile(FileID)
    ElseIf ParamInfo\ErrStatus = #PBIO_OK
      ParamInfo\ErrStatus = #PBIO_ErrCopyDriver
    EndIf
    
  Else
    FileID=2
  EndIf
  
  If FileID
    
    Driver_Install_M1:
    Mgr = OpenSCManager_(#Null, #Null, #SC_MANAGER_ALL_ACCESS)
    If Mgr
      Ser = CreateService_(Mgr, @Drivers_Name, @Drivers_Name, #SERVICE_ALL_ACCESS, #SERVICE_KERNEL_DRIVER, #SERVICE_SYSTEM_START, #SERVICE_ERROR_NORMAL, @File, #Null, 0, 0, 0, 0)
      ErrorCode=GetLastError_()
      If Ser 
        Result=#True 
      ElseIf ErrorCode=1073
        Result=#True 
      ElseIf ErrorCode=1072
        If Count=0
          
          If Ser : CloseServiceHandle_(Ser) : EndIf
          If Mgr : CloseServiceHandle_(Mgr) : EndIf
          dpUnInstall()
          Count+1
          Goto Driver_Install_M1
        EndIf
      EndIf
    EndIf
    
  EndIf
  
  If Ser : CloseServiceHandle_(Ser) : EndIf
  If Mgr : CloseServiceHandle_(Mgr) : EndIf
  
  If Result=#False And ParamInfo\ErrStatus = #PBIO_OK
    ParamInfo\ErrStatus = #PBIO_ErrInstallDriver
  EndIf
  
  If ParamInfo\ErrStatus = #PBIO_OK
    Result=Driver_Start()
  EndIf
  
  ProcedureReturn Result
EndProcedure






DataSection
  Driver:
  IncludeBinary "Driver\DriverPortIO.sys"
  DriverEnd:
  
  Driver_x64:
  IncludeBinary "Driver\DriverPortIO_x64.sys"
  DriverEnd_x64:
EndDataSection

EndModule

; IDE Options = PureBasic 5.31 (Windows - x86)
; Folding = ---
; EnableUnicode
; EnableXP